//---------------------------------------------------------------------------


#pragma hdrstop

#include "uRunningLine.h"

//---------------------------------------------------------------------------

#pragma package(smart_init)
//-----------------------
TPoint fnGetTriPt(TPoint ps, TPoint pe, float fAng);
_Horn fnGetHorn(TPoint ps, TPoint pe, float fAng);
void fnGetSnow(TPoint *Pt, int &iPtCnt, float fAng, int &iLv, bool bMult);
void fn2DCoorRad(double du, double dv, double dDeg, double &dx, double &dy);
TPoint fnRadPt(TPoint pt, float fAng);
///---------------------------------------------------------------------------
void fn2DCoorRad(double du, double dv, double dDeg, double &dx, double &dy)
{
   dx = du*cos(dDeg*M_PI/180.) - dv*sin(dDeg*M_PI/180.);
   dy = du*sin(dDeg*M_PI/180.) + dv*cos(dDeg*M_PI/180.);
}
//------------------------------------------------------------------------------
TPoint fnRadPt(TPoint pt, float fAng)
{
   double du, dv, dx, dy;
   du = pt.x;
   dv = pt.y;
   fn2DCoorRad(du, dv, fAng, dx, dy);
   return TPoint(dx, dy);
}
//----------------------------------------------------
void fnGetSnow(TPoint *Pt, int &iPtCnt, float fAng, int &iLv, bool bMult)
{
   int iCh;
   int iCnt=0;
   int iNewCnt=0;
   _Horn *pHorn;
   TPoint *ptTmp;
   TPoint ptParm = TPoint(1,1);
   ptTmp = new TPoint[iPtCnt];
   memcpy((void *)ptTmp, (void *)Pt, iPtCnt*sizeof(TPoint));
   iCh  = bMult ? -3 : 3;

  if(iPtCnt <iCh){
      pHorn = new _Horn[1];
      pHorn[0] = fnGetHorn(Pt[0], Pt[1], fAng);
      Pt[iNewCnt++] = pHorn[0].p0;
      Pt[iNewCnt++] = pHorn[0].p1;
      Pt[iNewCnt++] = pHorn[0].p2;
   }
   else{
      pHorn = new _Horn[iPtCnt];
      for (; iCnt<iPtCnt-1 ; iCnt++)
         pHorn[iCnt] = fnGetHorn(Pt[iCnt], Pt[iCnt+1], fAng);
      pHorn[iCnt] = fnGetHorn(Pt[iCnt], Pt[0], fAng);
      for (int i=0 ; i<iPtCnt ; i++){
         Pt[iNewCnt++] = ptTmp[i] ;
         Pt[iNewCnt++] = pHorn[i].p0;
         Pt[iNewCnt++] = pHorn[i].p1;
         Pt[iNewCnt++] = pHorn[i].p2;
      }
   }
   iPtCnt = iNewCnt;
   delete [] pHorn;
   delete [] ptTmp;
   if(--iLv)
      fnGetSnow(Pt, iPtCnt,  fAng, iLv, bMult);
}
//-----------------------------------------------------------------------------
_Horn fnGetHorn(TPoint ps, TPoint pe, float fAng)
{
   _Horn Result;
   double dDom = (1./3.);
   double dLenM, ddx, ddy, dsx, dsy, dex, dey ;
   dsx = ps.x; dsy = ps.y;
   dex = pe.x; dey = pe.y;
   ddx = dex - dsx;
   ddy = dey - dsy;
   dLenM = sqrt(ddx*ddx + ddy*ddy);
   if(ddy==0){
      Result.p0 = TPoint(dsx + ddx*dDom, dsy);
      Result.p2 = TPoint(dsx + ddx*(1.0-dDom), dsy);
   }
   else if(ddx==0){
      Result.p0 = TPoint(dsx, dsy + ddy*dDom);
      Result.p2 = TPoint(dsx, dsy + ddy*(1.0-dDom));
   }
   else{
      Result.p0.x = dsx + ddx*dDom;
      Result.p2.x = dsx + ddx*(1.0-dDom);
      Result.p0.y = dsy + ddy*dDom;
      Result.p2.y = dsy + ddy*(1.0-dDom);
   }
   Result.p1 = fnGetTriPt(Result.p0, Result.p2, fAng);
   return Result;
}
//-------------------------------------------------------------------
TPoint fnGetTriPt(TPoint ps, TPoint pe, float fAng)
{
   double dClkWs = 1.0; //clockwise=1.0 , Counter-clockwise=-1;
   double dM, dLenM, dsx, dsy, dex, dey, ddx, ddy;
   TPoint Result;
   fAng+=1;
   dsx = ps.x; dsy = ps.y;
   dex = pe.x; dey = pe.y;
   ddx = dex - dsx;
   ddy = dey - dsy;
   TPoint pm = TPoint(0.5*(dsx + dex), 0.5*(dsy + dey));
   if(dsy != dey)
      dM = -ddx/ddy;
   else
      dM = (ddx >= 0) ? NegInfinity : Infinity;

   dLenM = 0.5*sqrt(ddx*ddx + ddy*ddy)*tan(fAng*M_PI/180.);
   if(IsInfinite(dM)){
      if(ddx >= 0)
         Result = TPoint(pm.x, pm.y - (dClkWs*dLenM));
      else
         Result = TPoint(pm.x, pm.y + (dClkWs*dLenM));
   }
   else if(dM !=0){
      if(ddx >0 && ddy >0){
         Result.x = pm.x + dClkWs*sqrt(dLenM*dLenM/(dM*dM +1));
         Result.y = pm.y - dClkWs*sqrt(dLenM*dLenM/((1./dM*dM) +1));
      }
      else if(ddx <0 && ddy >0){
         Result.x = pm.x + dClkWs*sqrt(dLenM*dLenM/(dM*dM +1));
         Result.y = pm.y + dClkWs*sqrt(dLenM*dLenM/((1./dM*dM) +1));
      }
      else if(ddx <0 && ddy <0){
         Result.x = pm.x - dClkWs*sqrt(dLenM*dLenM/(dM*dM +1));
         Result.y = pm.y + dClkWs*sqrt(dLenM*dLenM/((1./dM*dM) +1));
      }
      else{
         Result.x = pm.x - dClkWs*sqrt(dLenM*dLenM/(dM*dM +1));
         Result.y = pm.y - dClkWs*sqrt(dLenM*dLenM/((1./dM*dM) +1));
      }
   }
   else{
      if(ddy >0 ){
         Result.x = pm.x + dClkWs*dLenM;
         Result.y = pm.y;
      }
      else{
         Result.x = pm.x - dClkWs*dLenM;
         Result.y = pm.y;
      }
   }
   return Result;
}
//------------------------------------------------------
__fastcall _RunningLine::_RunningLine(HWND hWnd)
           :TWinControl(hWnd)
{
   m_Timer = new TTimer(this);
   m_hWnd = hWnd;
   m_cvs = new TCanvas();
   m_hDC =GetDC(m_hWnd);
   m_cvs->Handle=m_hDC;
   m_Mem = new TMemoryStream();
   m_Timer->Enabled=false;
   m_Timer->OnTimer=fnOnTime;
   m_iPhase=0;
   m_clSlc=0;
   m_aCl[0] = clRed;
   m_aCl[1] = clYellow;
   m_aCl[2] = clBlue;
   fnSetInterval(250);
   fnSetElpLen(55);
   fnSetLineWidth(1);
}
//-------------------------------
__fastcall _RunningLine::~_RunningLine()
{
  ReleaseDC(m_hWnd, m_hDC);
  delete m_cvs;
  delete m_Timer;
  delete m_Mem;
}
//----------------------
void _RunningLine::fnSetInterval(int iMsec)
{
   m_Timer->Interval=iMsec;
}
//----------------------
void _RunningLine::fnSetLineWidth(int iWidth)
{
   m_iLineWidth=iWidth;
}
//------------------------
void _RunningLine::fnSetElpLen(int iLen)
{
  m_iElpLen=iLen;
}
//----------------------
void __fastcall _RunningLine::fnOnTime(TObject *Sender)
{
   fnDraw();
}
//--------------------------------------------
void _RunningLine::fnSetDrawPt(TPoint *ppt, int iNum)
{
   m_Mem->Clear();
   m_Mem->WriteBuffer((void *)ppt, iNum*sizeof(TPoint));
}
//-------------------------------
void _RunningLine::fnSwop(int &ix, int &iy)
{
   int iTmp;
   iTmp = ix;
   ix = iy;
   iy = iTmp;
}
//------------------------
int _RunningLine::fnDxPix(TPoint Pt0, TPoint Pt1, int iPicOfx, const int iElpLen)
{
   float fSlop=0;
   int x, y, dx, dy, iReadLen;
   dx = Pt1.x-Pt0.x;
   dy = Pt1.y-Pt0.y;
   int iElpCnt=iPicOfx;
   float fA,fx,fy, fk;
   if(dx){
      fSlop = dy/(float)dx;
 //     iElpCnt=iPicOfx * cos(atan(fSlop));
      for(int i=0;i<abs(dx) ;i++){
         if(dx >0)  x = Pt0.x + i;
         else       x = Pt0.x - i;
         if(dy >=0) y = Pt0.y + fabs(fSlop)*i;
         else       y = Pt0.y - fabs(fSlop)*i;
         if(!iElpCnt--){
            m_clSlc = (m_iPhase++)%cclNum;
            if(fSlop) iElpCnt = abs(iElpLen* cos(atan(fSlop)));
            else      iElpCnt = iElpLen;
         }

         for (int j=0;j<m_iLineWidth; j++){
            if(fSlop){
               fA = atan(-1/fSlop);
               fx = j*cos(fA);
               fy = j*sin(fA);
               m_cvs->Pixels[x+fx][y+fy]=m_aCl[m_clSlc];
            }
            else  m_cvs->Pixels[x][y+j]=m_aCl[m_clSlc];
         }
      }
   }
   else{
      for(int i=0;i<abs(dy) ;i++){
         x = Pt0.x;
         if(dy>0)      y = Pt0.y + i;
         else if(dy<0) y = Pt0.y - i;
         else          y = Pt0.y;
         if(!iElpCnt--){
            m_clSlc = (m_iPhase++)%cclNum;
            iElpCnt=iElpLen;
         }
         for (int j=0;j<m_iLineWidth; j++)
            m_cvs->Pixels[x+j][y]=m_aCl[m_clSlc];
      }
   }
//   if(fSlop)
 //     iElpCnt = abs(iElpCnt / cos(atan(fSlop)));
   return iElpCnt;
}
//------------------------
int _RunningLine::fnDyPix(TPoint Pt0, TPoint Pt1, int iPicOfx, const int iElpLen)
{
   float fInvSlop;
   int x, y, dx, dy, iReadLen;
   dx = Pt1.x-Pt0.x;
   dy = Pt1.y-Pt0.y;
   int iElpCnt=iPicOfx;
   float fA,fx,fy;
   if(dy){
      fInvSlop = dx/(float)dy;
 //     iElpCnt=iPicOfx * cos(atan(fInvSlop));
      for(int i=0;i<abs(dy) ;i++){
         if(dy >0)  y = Pt0.y + i;
         else       y = Pt0.y - i;
         if(dx >=0) x = Pt0.x + fabs(fInvSlop)*i;
         else       x = Pt0.x - fabs(fInvSlop)*i;
         if(!iElpCnt--){
            m_clSlc = (m_iPhase++)%cclNum;
            if(fInvSlop) iElpCnt = abs(iElpLen* cos(atan(fInvSlop)));
            else         iElpCnt = iElpLen;
         }
       //  fA = atan(fInvSlop);
         for (int j=0;j<m_iLineWidth; j++){
            if(fInvSlop){
               fA = atan(-1/fInvSlop);
               fx = j*cos(fA);
               fy = j*sin(fA);
               m_cvs->Pixels[x+fx][y+fy]=m_aCl[m_clSlc];
            }
            else  m_cvs->Pixels[x+j][y]=m_aCl[m_clSlc];
         }
      }
   }
   else{
      for(int i=0;i<abs(dx) ;i++){
         y = Pt0.y;
         if(dx>0)      x = Pt0.x + i;
         else if(dy<0) x = Pt0.x - i;
         else          x = Pt0.x;
         if(!iElpCnt--){
            m_clSlc = (m_iPhase++)%cclNum;
            iElpCnt=iElpLen;
         }
         for (int j=0;j<m_iLineWidth; j++)
            m_cvs->Pixels[x][y+j]=m_aCl[m_clSlc];
      }
   }
  // if(fInvSlop)
  //    iElpCnt =  abs(iElpCnt / cos(atan(fInvSlop)));
   return iElpCnt;
}
//-------------------------
void _RunningLine::fnDraw()
{
   int dx, dy, iReadLen;
   TPoint Pt0, Pt1;
   m_Mem->Position=0;
   static int iOfx=0;
   if(m_Mem->Size){
     m_Mem->Read((void *)&Pt0, sizeof(TPoint));
      do{
         iReadLen=m_Mem->Read((void *)&Pt1, sizeof(TPoint));
         if(iReadLen == sizeof(TPoint)){
            dx = Pt1.x-Pt0.x;
            dy = Pt1.y-Pt0.y;
            if(abs(dx) >= abs(dy))
              iOfx = fnDxPix(Pt1, Pt0, iOfx, m_iElpLen);
            else
              iOfx = fnDyPix(Pt1, Pt0, iOfx, m_iElpLen);
         }
         Pt0=Pt1;
      }while(iReadLen == sizeof(TPoint));
   }
}
//---------------------------------
void _RunningLine::fnRunning()
{
   m_Timer->Enabled=true;
}
void _RunningLine::fnStop()
{
   m_Timer->Enabled=false;
}

